
!********************************************************************************
!*   Copyright 1999 -- 2002 C. Kind, M. Reiher, J. Neugebauer, and B. A. Hess,
!*   Chair of Theoretical Chemistry, Friedrich-Alexander-University 
!*   Erlangen-Nuremberg, Egerlandstr. 3, D-91058 Erlangen, Germany.
!*
!*
!*   This file is part of the package SNF.
!*
!*   SNF is free software; you can redistribute it and/or modify
!*   it under the terms of the GNU General Public License as published by
!*   the Free Software Foundation; either version 2 of the License, or
!*   (at your option) any later version.
!*
!*   This program is distributed in the hope that it will be useful,
!*   but WITHOUT ANY WARRANTY; without even the implied warranty of
!*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!*   GNU General Public License for more details.
!*
!*   You should have received a copy of the GNU General Public License
!*   along with this program; if not, write to the Free Software
!*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!*
!********************************************************************************
module snf_hessian
  use f90_kind
  use snf_error

  private
  public :: create_hessian_xyz, print_hessian
  public :: create_hessian_xyz_fb
  public :: destroy_hessian_xyz,expand_hessian, check_hessian
  public :: balance_hessian,symmetrize_hessian, hessian_mass_weighting
  !************************************************************************


contains

  !************************************************************************

  subroutine create_hessian_xyz(hessian_xyz,gradients,distortion)
    implicit none
    real(double),intent(in) ::gradients(:,:,:),distortion(:)
    real(double), pointer :: hessian_xyz(:,:)
    integer :: i,j,nvar,idum
    !----------------------------------------------------------------
    ! calculation of second derivative of energy with following formula:
    ! 
    !  d^2e      de      de  
    ! ------ =   ---  -  --- 
    ! dx1dxi     dxi     dxi 
    !           --------------
    !           2*distortion(x1)
    !
    ! unit is hartree/bohr^2
    !
    ! written by Carsten Kind and Markus Reiher, Uni Erlangen, 1999
    ! modified by J. Neugebauer, Uni Erlangen, 2001 
    ! (correction of 5-point formula)
    ! extended by J. Neugebauer, Uni Erlangen, 2002 
    ! (introduction of 7-point formula)
    !----------------------------------------------------------------
    nvar=ubound(gradients,1)
    idum=ubound(gradients,3)
    allocate(hessian_xyz(nvar,nvar))
    if (idum.eq.2) then
       do i=1,nvar
          do j=1,nvar
             hessian_xyz(j,i)=0.5d0/distortion(i)* &
                  &           (gradients(j,i,1)-gradients(j,i,2))
          enddo
       enddo
    else if (idum.eq.4) then
       do i=1,nvar
          do j=1,nvar
             hessian_xyz(j,i)=1.0d0/(24.0d0*distortion(i))* &
                  &           ( -2.0d0*gradients(j,i,1) &
                  &            +16.0d0*gradients(j,i,2) &
                  &            -16.0d0*gradients(j,i,3) &
                  &             +2.0d0*gradients(j,i,4))
          enddo
       enddo
    else if (idum.eq.6) then
       do i=1,nvar
          do j=1,nvar
             hessian_xyz(j,i)=1.0d0/(720.0d0*distortion(i))* &
                  &           ( 12.0d0*gradients(j,i,1) &
                  &           -108.0d0*gradients(j,i,2) &
                  &           +540.0d0*gradients(j,i,3) &
                  &           -540.0d0*gradients(j,i,4) &
                  &           +108.0d0*gradients(j,i,5) &
                  &            -12.0d0*gradients(j,i,6))
          enddo
       enddo
    else
       call serror('create_hessian_xyz',2,'Wrong ctrl%numderiv value!')
    end if
  end subroutine create_hessian_xyz

  !************************************************************************

  subroutine create_hessian_xyz_fb(hessian_xyz,gradients,distortion,fbtype)
    use snf_rgrad,only : read_gradient
    implicit none
    real(double),intent(in) ::gradients(:,:,:),distortion(:)
    real(double), pointer :: hessian_xyz(:,:),gradeq(:)
    real(double) :: escf
    integer :: i,j,nvar,idum,info
    character(len=5) :: fbtype
    !----------------------------------------------------------------
    ! calculation of second derivative of energy with forward/backward
    ! differences
    ! 
    ! unit is hartree/bohr^2
    !
    ! NOTE: THIS IS A TESTMODULE; IN PARTICULAR, IT IS ASSUMED THAT
    !       THE COMPONENTS OF THE GRADIENT FOR THE EQUILIBRIUM STRUCTURE
    !       ARE 0.0D0, WHICH IS CERTAINLY NOT TRUE.
    !
    !       FOR BETTER PERFORMANCE, THE GRADIENT OF THE EQUILIBRIUM
    !       STRUCTURE SHOULD BE READ AS WELL.
    !
    ! written by J. Neugebauer, Uni Erlangen, 2002 
    !----------------------------------------------------------------

    nvar=ubound(gradients,1)
!===============================================
! try to read gradient for equilibrium structure
!===============================================
    allocate(gradeq(nvar))
    call read_gradient(escf,gradeq,info)
    if(info.ne.0) then
      call serror('create_hessian_xyz_fb',0,'Could not read gradients' &
    & //' for equilibrium structure !## Zero gradient is assumed instead !',6)
    end if
    write(*,*) 'gradients equilibrium structure:'
    do i=1,nvar
      write(*,*) i,gradeq(i)
    end do
    idum=ubound(gradients,3)
    allocate(hessian_xyz(nvar,nvar))
    if (idum.eq.2) then
      if(fbtype.eq.'fward') then
        do i=1,nvar
          do j=1,nvar
            hessian_xyz(j,i)=(gradients(j,i,1)-gradeq(j))/distortion(i)
          enddo
        enddo
      end if
      if(fbtype.eq.'bward') then
        do i=1,nvar
          do j=1,nvar
            hessian_xyz(j,i)=(gradeq(j)-gradients(j,i,2))/distortion(i)
          enddo
        enddo
      end if
    else if (idum.eq.4) then
      if(fbtype.eq.'fward') then
        do i=1,nvar
          do j=1,nvar
            hessian_xyz(j,i)=(gradients(j,i,1)-gradients(j,i,2))/distortion(i)
          enddo
        enddo
      else if(fbtype.eq.'bward') then
        do i=1,nvar
          do j=1,nvar
            hessian_xyz(j,i)=(gradients(j,i,3)-gradients(j,i,4))/distortion(i)
          enddo
        enddo
      end if
    else
       call serror('create_hessian_xyz',2,'Wrong ctrl%numderiv value!')
    end if
    deallocate(gradeq)
  end subroutine create_hessian_xyz_fb

  !************************************************************************

  subroutine print_hessian(hessian_xyz,filename)
    !----------------------------------------------------------------
    !     output routine
    !----------------------------------------------------------------
    implicit none
    real(double),intent(in) :: hessian_xyz(:,:)
    character*(*) filename
    integer :: i,j,nvar
    nvar=ubound(hessian_xyz,1)
    open(unit=45,file=filename,status='unknown')
    write(45,*) &
         & 'hessian matrix in cartesian coordinates'
    write(45,*)'nvar=',nvar    
    write(45,'(4d18.10)')((hessian_xyz(i,j),j=1,i),i=1,nvar)
    close (45)
  end subroutine print_hessian

  !************************************************************************

  subroutine expand_hessian(hessian_xyz,rep,symm_index)
    use snf_symmat,only : symel_mat
    use snf_tools
    !----------------------------------------------------------------
    ! copy hessian calculated on irreducible atoms to the 
    ! reducible atoms. this is done by multiplying symmetry representations in 
    ! the derivative basis from the left and from the right with the dipol 
    ! moment derivates and selecting only the required columns.
    !
    ! written by Carsten Kind, Uni Erlangen, 1999
    !----------------------------------------------------------------
    implicit none
    real(double),pointer :: hessian_xyz(:,:),rot_hessian(:,:),expanded_hessian(:,:),tmpmat(:,:)
    real(double),intent(in) :: rep(:,:,:)
    integer, intent(in) :: symm_index(:)
    integer :: s,i,offset,nvar,nsymel
    nsymel=ubound(rep,3)
    nvar=ubound(hessian_xyz,2)
    allocate(rot_hessian(nvar,nvar),expanded_hessian(nvar,nvar),tmpmat(nvar,nvar))
    do s=1,nsymel

!-------------------------------------------------------------
! matrix multiplication might be included in if-statement ?!
! (JN, 08/2001)
!-------------------------------------------------------------

       call dgemm('N','N',nvar,nvar,nvar,1.D0,hessian_xyz(1,1),nvar,rep(1,1,s),nvar,0.D0,tmpmat(1,1),nvar)
       call dgemm('T','N',nvar,nvar,nvar,1.D0,rep(1,1,s),nvar,tmpmat(1,1),nvar,0.D0,rot_hessian(1,1),nvar)
       do i=1,ubound(symm_index,1)
          if(symm_index(i) .eq. s) then
             offset=(i-1)*3
             expanded_hessian(:,offset+1:offset+3)=rot_hessian(:,offset+1:offset+3)
          endif
       enddo
    enddo
    deallocate(hessian_xyz,rot_hessian,tmpmat)
    hessian_xyz => expanded_hessian
  end subroutine expand_hessian

  !************************************************************************

  subroutine symmetrize_hessian(h_xyz,rep)
    !----------------------------------------------------------------
    ! written by Carsten Kind, Uni Erlangen, 1999
    ! symmetry check included 
    ! by Johannes Neugebauer, Uni Erlangen, 2001
    !----------------------------------------------------------------
    use snf_tools
    implicit none
    real(double),intent(inout) :: h_xyz(:,:)
    real(double) :: fac,errsum
    real(double),intent(in) :: rep(:,:,:)
    real(double),pointer :: h_rot(:,:),tmpmat(:,:),tmpmat2(:,:)
    integer :: s,nvar,nsymel,i,j
    character(len=60) :: errbuff
    nvar=ubound(h_xyz,2)
    nsymel=ubound(rep,3)
    fac=1.D0/dble(nsymel)
    allocate(h_rot(nvar,nvar),tmpmat(nvar,nvar),tmpmat2(nvar,nvar))
    h_rot=0.0D0
    do s=1,nsymel
       call dgemm('N','N',nvar,nvar,nvar,1.D0,h_xyz(1,1),nvar,rep(1,1,s),nvar,0.D0,tmpmat(1,1),nvar)
!CJN >>>>>>>>>>>>>
       tmpmat2=0.0d0
!       call dgemm('T','N',nvar,nvar,nvar,1.D0,rep(1,1,s),nvar,tmpmat(1,1),nvar,1.D0,h_rot(1,1),nvar)
       call dgemm('T','N',nvar,nvar,nvar,1.D0,rep(1,1,s),nvar,tmpmat(1,1),nvar,0.D0,tmpmat2(1,1),nvar)
       errsum=0.0d0
       do i=1,nvar
         do j=1,nvar
           errsum=errsum+abs(h_xyz(i,j)-tmpmat2(i,j))
         end do
       end do
       errsum=errsum/dble(nvar)/dble(nvar)
       if(errsum.gt.1.0d-3) then
         write(errbuff,'(a,f8.4,a,i3)') 'average error = ',errsum, ' for s = ',s
         call serror('symmetrize_hessian',0,trim(adjustl(errbuff)))
       end if
       h_rot=h_rot + tmpmat2
!CJN <<<<<<<<<<<<<
    enddo
    h_xyz=fac*h_rot
    deallocate(h_rot,tmpmat,tmpmat2)
  end subroutine symmetrize_hessian

  !************************************************************************

  subroutine destroy_hessian_xyz(hessian_xyz)
    implicit none
    !----------------------------------------------------------------
    ! clean up memory
    !
    ! written by Carsten Kind, Uni Erlangen, 1999
    !----------------------------------------------------------------
    real(double), pointer :: hessian_xyz(:,:)
    if(associated(hessian_xyz)) deallocate(hessian_xyz)
  end subroutine destroy_hessian_xyz

  !************************************************************************

  subroutine check_hessian(hessian_xyz,ctrl,symm_index,istepfl,new_restart,unit)
    !----------------------------------------------------------------
    ! Determine which atoms are to be recalculated based on the asymmetry of 
    ! the Hessian matrix.
    !
    ! written by Carsten Kind, Uni Erlangen, 1999
    !----------------------------------------------------------------
    use snf_control, only: type_ctrl
    implicit none

    type type_error
       integer :: index,number
       real(double) :: value
    end type type_error

    type(type_ctrl) :: ctrl      ! essential control flags and parameters
    real(double),pointer :: hessian_xyz(:,:)
    integer,optional :: unit
    character(len=3) :: xyz
    integer :: i,j,nvar,atom_i,xyz_i,max_diff,index_max_diff,out
    type(type_error) :: tmp_error
    type(type_error),pointer :: error_vec(:)
    real(double) :: threshold,maxerr,meanerr,err
    integer, pointer :: istepfl(:),symm_index(:)
    logical,intent(out) :: new_restart
    if(present(unit)) then
       out=unit
    else
       out=6
    endif
    new_restart = .false.
    nvar=ubound(hessian_xyz,1)
    allocate(error_vec(nvar))
    xyz='xyz'
    threshold = 6.d01 * ctrl%cstep**2
    !----------------------------------------------------------------
    ! count all differences between h(i,j) and h(j,i) larger than 100*cstep**2
    !----------------------------------------------------------------
    maxerr=0.0d0
    do i=1,nvar
       error_vec(i)%number=0
       error_vec(i)%value=0.d0
       do j=1,nvar
          err=abs(hessian_xyz(i,j)-hessian_xyz(j,i))
          if(err.gt.threshold) then
             error_vec(i)%value = error_vec(i)%value + err
             error_vec(i)%number=error_vec(i)%number+1
          endif
          maxerr=max(maxerr,err)
          meanerr=meanerr+err*err
       enddo
    enddo
    meanerr=sqrt(meanerr/(nvar*nvar-1))
    write(out,*)
    write(out,'('' Hessian symmetry check information   '')')
    write(out,'('' ------------------------------------ '')')
    write(out,'(''   Maximal absolute error |hij-hji| : '',G9.2)')maxerr
    write(out,'(''   Standard deviation               : '',G9.2)')meanerr
    !----------------------------------------------------------------
    ! save indices
    !----------------------------------------------------------------
    do i=1,nvar
       error_vec(i)%index=i
    enddo
    !----------------------------------------------------------------
    ! sort error-vector with decreasing order of error
    !----------------------------------------------------------------
    do i=1,nvar-1
       do j=i+1,nvar
          if(error_vec(i)%number .lt. error_vec(j)%number) then
             tmp_error = error_vec(i)
             error_vec(i) = error_vec(j)
             error_vec(j)=tmp_error
          endif
       enddo
    enddo
    if(error_vec(1)%number .gt. 0) then
       !----------------------------------------------------------------
       ! determine index of the vector with the largest decrease of 
       ! of errors according to the next vector
       !----------------------------------------------------------------
       write(out,*)
       write(out,'(''******************************************************'')')
       write(out,'(''ERROR: Hessian is not symmetric'')')
       write(out,*)
       write(out,'(''Error criterion: |hij-hji| > threshold='',f15.6)') &
            &  threshold
       write(out,*)
       write(out,'('' atom  | coord.|  error    | sum of |hij-hji|'')')
       write(out,'(''number | [xyz] |  counts   | for all errors.'')')
       write(out,'(''-------|-------|-----------|-----------------'')')
       do i=1,nvar
          atom_i=(error_vec(i)%index+2)/3
          xyz_i=error_vec(i)%index-(atom_i-1)*3
          write(out,'(i5,2x,''|   '',a1,''   |'',i6,''     |'',f15.6)') &
               &    atom_i,xyz(xyz_i:xyz_i),&
               &    error_vec(i)%number, error_vec(i)%value
       enddo
       write(out,'(''******************************************************'')')
       write(out,*)
       do i=2,nvar
          if((error_vec(i-1)%number-error_vec(i)%number).gt.max_diff) then
             max_diff = error_vec(i-1)%number-error_vec(i)%number
             index_max_diff=i-1
          endif
       enddo
       write(out,*)
       !----------------------------------------------------------------
       ! determine all irreducible vectors with the largest number of errors
       ! and print warning
       !----------------------------------------------------------------
       do i=1,index_max_diff
          if(symm_index((error_vec(i)%index+2)/3) .eq. 1) then
             atom_i=(error_vec(i)%index+2)/3
             xyz_i=error_vec(i)%index-(atom_i-1)*3
             istepfl(error_vec(i)%index*2-1) = 0
             istepfl(error_vec(i)%index*2) = 0
             write(out,'(''atom '',i3,'' coordinate '',a1,'' has to be recalculated'')') &
                  & atom_i,xyz(xyz_i:xyz_i)
             new_restart = .true.
          endif
       enddo
       write(out,*)
    endif
    deallocate(error_vec)
  end subroutine check_hessian

  !************************************************************************

  subroutine balance_hessian(hessian)
    !----------------------------------------------------------------
    ! symmetrize Hessian
    !
    ! written by Carsten Kind, Uni Erlangen, 1999
    !----------------------------------------------------------------
    implicit none
    real(double),intent(inout) :: hessian(:,:)
    integer i,j,nvar
    nvar=ubound(hessian,1)
    do j=1,nvar
       do i=1,j-1
          hessian(i,j)=0.5d0*(hessian(i,j)+hessian(j,i))
          hessian(j,i) = hessian(i,j)
       enddo
    enddo
  end subroutine balance_hessian

  !************************************************************************

  subroutine hessian_mass_weighting(hessian_xyz,hessian_m,molecule)
    use snf_molecule, only: type_atom
    implicit none
    type(type_atom),intent(in) :: molecule(:)
    real(double), intent(in) :: hessian_xyz(:,:)
    real(double), intent(out) :: hessian_m(:,:)
    integer :: i,j,nvar
    !----------------------------------------------------------------
    ! mass weighting of hessian
    !                      hessian_xyz(i,j)
    ! hessian_xyz(i,j) = -------------------
    !                    sqrt(m_i)*sqrt(m_j)
    !
    ! unit is hartree*mol/(bohr^2*g)
    !
    ! written by Carsten Kind, Uni Erlangen, 2000
    !----------------------------------------------------------------
    nvar=ubound(hessian_xyz,1)
    do j=1,nvar
       do i=1,j
          hessian_m(i,j)=hessian_xyz(i,j)&
               & /sqrt(molecule((i+2)/3)%mass*molecule((j+2)/3)%mass)
          hessian_m(j,i)=hessian_m(i,j)
       enddo
    enddo
  end subroutine hessian_mass_weighting
 
  !************************************************************************
                                              

end module snf_hessian
